#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec


class AcTransformersV1(BaseOpenEpdHierarchicalSpec):
    """Equipment for transforming between higher and lower voltage AC power."""

    _EXT_VERSION = "1.0"


class ElectricalInsulatorsV1(BaseOpenEpdHierarchicalSpec):
    """Passive electrical isolation."""

    _EXT_VERSION = "1.0"


class ElectricalSubstationsV1(BaseOpenEpdHierarchicalSpec):
    """Electrical substations performance specification."""

    _EXT_VERSION = "1.0"


class ElectricalSwitchgearV1(BaseOpenEpdHierarchicalSpec):
    """
    Equipment for interrupting and controlling high-power electrical flows.

    Used for protection, isolation, or control of electrical equipment.
    """

    _EXT_VERSION = "1.0"


class PowerCablingV1(BaseOpenEpdHierarchicalSpec):
    """High-voltage electrical cabling."""

    _EXT_VERSION = "1.0"


class ElectricalTransmissionAndDistributionEquipmentV1(BaseOpenEpdHierarchicalSpec):
    """Electrical Transmission & Distribution Equipment."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    AcTransformers: AcTransformersV1 | None = None
    ElectricalInsulators: ElectricalInsulatorsV1 | None = None
    ElectricalSubstations: ElectricalSubstationsV1 | None = None
    ElectricalSwitchgear: ElectricalSwitchgearV1 | None = None
    PowerCabling: PowerCablingV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "BuildingHeatingPipingRangeV1",
    "BuriedPipingRangeV1",
    "UtilityPipingRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeAmount
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import BuriedPipingType, PipingAnsiSchedule, UtilityPipingMaterial
from openepd.model.validation.quantity import AmountRangeLengthMm


class BuildingHeatingPipingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Heating piping.

    System of pipes used to supply heated fluids (liquids or steam) for purposes of controlling temperature inside a
    home, business, or other building facility.

    Range version.
    """

    _EXT_VERSION = "1.0"


class BuriedPipingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    System of pipes used to provide or transport fluids (liquids and gases) underground.

    Range version.
    """

    _EXT_VERSION = "1.0"

    buried_piping_type: list[BuriedPipingType] | None = pyd.Field(default=None, description="")


class UtilityPipingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Utility piping.

    System of pipes used to convey fluids (liquids and gases) from one location to another. Piping can be metal,
    plastic, concrete, fiberglass, or other materials.

    Range version.
    """

    _EXT_VERSION = "1.0"

    thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    piping_diameter: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    mass_per_unit_length: RangeAmount | None = pyd.Field(default=None, description="")
    piping_ansi_schedule: list[PipingAnsiSchedule] | None = pyd.Field(default=None, description="")
    utility_piping_material: list[UtilityPipingMaterial] | None = pyd.Field(default=None, description="")
    BuildingHeatingPiping: BuildingHeatingPipingRangeV1 | None = None
    BuriedPiping: BuriedPipingRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from enum import StrEnum
from typing import Literal

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseDocumentFactory, OpenEpdDoctypes
from openepd.model.common import WithAltIdsMixin, WithAttachmentsMixin
from openepd.model.declaration import AverageDatasetMixin, BaseDeclaration, RefBase
from openepd.model.lcia import WithLciaMixin
from openepd.model.org import Org
from openepd.model.versioning import OpenEpdVersions, Version


class LicenseTerms(StrEnum):
    """Licensing terms."""

    CC_BY = "CC-BY"
    """
    Creative Commons attribution-only license https://creativecommons.org/licenses/by/4.0/.
    """
    ODbL = "ODbL"
    """
    Open Database License per https://opendatacommons.org/licenses/odbl/
    """
    Government = "Government"
    """
    The data is offered at no charge by a government body such as the USDA or Okobaudat.
    """
    Other = "Other"
    """
    Review estimate documentation to determine license terms.
    """


class GenericEstimateRef(RefBase, title="Generic Estimate (Ref)"):
    """Reference (short) version of Generic Estimate object."""


class GenericEstimatePreviewV0(
    WithAttachmentsMixin, AverageDatasetMixin, GenericEstimateRef, BaseDeclaration, title="Generic Estimate (preview)"
):
    """
    Generic Estimate preview, used in API list responses and where there is no need for a full object.

    Excludes LCIA data.
    """

    _FORMAT_VERSION = OpenEpdVersions.Version0.as_str()

    doctype: Literal["openGenericEstimate"] = pyd.Field(
        description='Describes the type and schema of the document. Must always be "openGenericEstimate"',
        default="openGenericEstimate",
    )

    publisher: Org | None = pyd.Field(description="Organization that published the LCA results.")
    reviewer: Org | None = pyd.Field(description="Org that performed a critical review of the LCA.")
    reviewer_email: pyd.EmailStr | None = pyd.Field(
        description="Email address of the third party verifier", example="john.doe@example.com", default=None
    )

    license_terms: LicenseTerms | None = pyd.Field(description="The license terms for use of the data.")
    model_repository: pyd.AnyUrl | None = pyd.Field(
        default=None, description="A link to the shared git repository containing the LCA model used for this estimate."
    )


GenericEstimatePreview = GenericEstimatePreviewV0


class GenericEstimateV0(GenericEstimatePreviewV0, WithLciaMixin, WithAltIdsMixin, title="Generic Estimate (Full)"):
    """Full Generic Estimate object."""


GenericEstimate = GenericEstimateV0


class GenericEstimateWithDepsV0(GenericEstimateV0, title="Generic Estimate (with Dependencies)"):
    """
    Expanded version of the GenericEstimate.

    Contains related entities - orgs - with full fields, to support object matching in implementations.
    """

    publisher: Org | None = pyd.Field(description="Organization that published the LCA results.")
    reviewer: Org | None = pyd.Field(description="Org that performed a critical review of the LCA.")


GenericEstimateWithDeps = GenericEstimateWithDepsV0


class GenericEstimateFactory(BaseDocumentFactory[GenericEstimate]):
    """Factory for EPD objects."""

    DOCTYPE_CONSTRAINT = OpenEpdDoctypes.GenericEstimate
    VERSION_MAP: dict[Version, type[GenericEstimate]] = {OpenEpdVersions.Version0: GenericEstimateV0}



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Annotated

from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec, CodegenSpec
from openepd.model.specs.enums import CmuBlockType, CmuWeightClassification
from openepd.model.validation.quantity import GwpKgCo2eStr, PressureMPaStr


class CMUV1(BaseOpenEpdHierarchicalSpec):
    """Pre-manufactured concrete masonry blocks."""

    _EXT_VERSION = "1.0"

    # Own fields:
    white_cement: bool | None = pyd.Field(default=None, description="", example=True)
    strength_28d: PressureMPaStr | None = pyd.Field(default=None, description="", example="1 MPa")
    weight_classification: CmuWeightClassification | None = pyd.Field(default=None, description="", example="Normal")
    block_type: CmuBlockType | None = pyd.Field(default=None, description="", example="Gray")
    insulated: bool | None = pyd.Field(default=None, description="", example=True)
    sound_performance: bool | None = pyd.Field(default=None, description="", example=True)
    b1_recarbonation: Annotated[GwpKgCo2eStr | None, CodegenSpec(override_type=GwpKgCo2eStr)] = pyd.Field(
        default=None, description="", example="1 kgCO2e"
    )
    b1_recarbonation_z: Annotated[float | None, CodegenSpec(override_type=float)] = pyd.Field(
        default=None, description="", example=2.3
    )



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from enum import StrEnum
from typing import Any, ClassVar

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema
from openepd.model.common import Measurement
from openepd.model.validation.quantity import ExternalValidationConfig


class EolScenario(BaseOpenEpdSchema):
    """
    A scenario for end-of-life.

    If a particular implementation commits to a specific scenario, then that value can be used;
    otherwise the outcomes can be statistically combined.
    """

    name: str = pyd.Field(
        max_length=40,
        example="Landfill",
        description="A brief text description of the scenario, preferably from list eol_scenario_names",
    )
    likelihood: float | None = pyd.Field(
        description="The weigting of this scenario used in the C1 .. C4 dataset. For example, the overall C1 shoudl be "
        "equal to weighted sum of C1 from all the scenarios, weighted by likelihood.",
        example=0.33,
        default=None,
    )
    C1: Measurement | None = pyd.Field(
        description="Deconstruction and Demolition under this scenario",
        default=None,
    )
    C2: Measurement | None = pyd.Field(
        description="Transport to waste processing or disposal under this scenario.",
        default=None,
    )
    C3: Measurement | None = pyd.Field(
        description="Waste Processing under this scenario",
        default=None,
    )
    C4: Measurement | None = pyd.Field(
        description="Disposal under this scenario",
        default=None,
    )
    D: Measurement | None = pyd.Field(
        description="Potential net benefits from reuse, recycling, and/or energy recovery beyond "
        "the system boundary under this scenario",
        default=None,
    )


class ScopeSet(BaseOpenEpdSchema):
    """
    A set of scopes, such as A1..A5, B1..B7, C1..C4, D, etc.

    Any scopes with no entry are considered 'measure not declared' (null) rather than zero.
    The 'unit' field must be consistent across all scopes in a single scopeset.
    """

    allowed_units: ClassVar[str | tuple[str, ...] | None] = None

    A1A2A3: Measurement | None = pyd.Field(
        description="Sum of A1..A3",
        default=None,
    )
    A1: Measurement | None = pyd.Field(
        description="Raw Material Supply",
        default=None,
    )
    A2: Measurement | None = pyd.Field(
        description="Transport to Manufacturing",
        default=None,
    )
    A3: Measurement | None = pyd.Field(
        description="Manufacturing",
        default=None,
    )
    A4: Measurement | None = pyd.Field(
        description="Transport to Construction",
        default=None,
    )
    A5: Measurement | None = pyd.Field(
        description="Construction",
        default=None,
    )
    B1: Measurement | None = pyd.Field(
        description="Use impacts over Reference Service Life (Predicted)",
        default=None,
    )
    B1_years: float | None = pyd.Field(
        gt=0,
        lt=100,
        description="Timeframe over which B1 is evaluated, in years. "
        "For example, an impact of 1.23 kgCO2e per year for ten years could be B1=1.23kgCO2e, "
        "B1_years=1.0 or  B1=12.3kgCO2e, B1_years=10.0",
        default=None,
    )
    B2: Measurement | None = pyd.Field(
        description="Predicted Maintenance Impacts over Reference Service Life",
        default=None,
    )
    B2_years: float | None = pyd.Field(
        gt=0,
        lt=100,
        description="Predicted Maintenance Impacts over Reference Service Life",
        default=None,
    )
    B3: Measurement | None = pyd.Field(
        description="Predicted Repair impacts over Reference Service Life",
        default=None,
    )
    B3_years: float | None = pyd.Field(
        gt=0,
        lt=100,
        description="Timeframe over which B3 is evaluated, in years",
        default=None,
    )
    B4: Measurement | None = pyd.Field(
        description="Predicted Replacement Impacts over the Building lifetime "
        "('Estimated Construction Works lifespan') specified in the PCR.",
        default=None,
    )
    B4_years: float | None = pyd.Field(
        gt=0,
        lt=100,
        description="Timeframe over which B4 is evaluated, in years",
        default=None,
    )
    B5: Measurement | None = pyd.Field(
        description="Predicted Refurbishment Impacts over the Building lifetime "
        "('Estimated Construction Works lifespan') specified in the PCR.",
        default=None,
    )
    B5_years: float | None = pyd.Field(
        gt=0,
        lt=100,
        description="Timeframe over which B5 is evaluated, in years",
        default=None,
    )
    B6: Measurement | None = pyd.Field(
        description="Predicted Impacts related to Operational Energy Use",
        default=None,
    )
    B6_years: float | None = pyd.Field(
        gt=0,
        lt=100,
        description="Timeframe over which B6 is evaluated, in years",
        default=None,
    )
    B7: Measurement | None = pyd.Field(
        description="Predicted Impacts related to Operational Water Use",
        default=None,
    )
    B7_years: float | None = pyd.Field(
        gt=0,
        lt=100,
        description="Timeframe over which B7 is evaluated, in years",
        default=None,
    )
    C_scenarios: list[EolScenario] | None = pyd.Field(
        description="A list of possible end-of-life scenarios, "
        "for use in analyses where the end-of-life can be predicted.",
        default=None,
    )
    C1: Measurement | None = pyd.Field(
        description="Deconstruction and Demolition",
        default=None,
    )
    C2: Measurement | None = pyd.Field(
        description="Transport to waste processing or disposal.",
        default=None,
    )
    C3: Measurement | None = pyd.Field(
        description="Waste Processing",
        default=None,
    )
    C4: Measurement | None = pyd.Field(
        description="Disposal",
        default=None,
    )
    D: Measurement | None = pyd.Field(
        default=None,
        description="Potential net benefits from reuse, recycling, and/or energy recovery beyond the system boundary.",
    )

    @pyd.root_validator
    def _unit_validator(cls, values: dict[str, Any]) -> dict[str, Any]:
        all_units = set()

        for k, v in values.items():
            if isinstance(v, Measurement):
                all_units.add(v.unit)

        if not cls.allowed_units:
            # For unknown units - only units should be the same across all measurements (textually)
            if len(all_units) > 1:
                raise ValueError("All scopes and measurements should be expressed in the same unit.")
        else:
            # might be multiple variations of the same unit (kgCFC-11e, kgCFC11e)
            if len(all_units) > 1 and ExternalValidationConfig.QUANTITY_VALIDATOR:
                all_units_list = list(all_units)
                first = all_units_list[0]
                for unit in all_units_list[1:]:
                    ExternalValidationConfig.QUANTITY_VALIDATOR.validate_same_dimensionality(first, unit)

        # can correctly validate unit
        if cls.allowed_units is not None and len(all_units) == 1 and ExternalValidationConfig.QUANTITY_VALIDATOR:
            unit = next(iter(all_units))
            allowed_units = cls.allowed_units if isinstance(cls.allowed_units, tuple) else (cls.allowed_units,)

            matched_unit = False
            for allowed_unit in allowed_units:
                try:
                    ExternalValidationConfig.QUANTITY_VALIDATOR.validate_same_dimensionality(unit, allowed_unit)
                    matched_unit = True
                except ValueError:
                    ...
            if not matched_unit:
                raise ValueError(
                    f"'{', '.join(allowed_units)}' is only allowed unit for this scopeset. Provided '{unit}'"
                )

        return values


class ScopesetByNameBase(BaseOpenEpdSchema, extra="allow"):
    """Base class for the data structures presented as typed name:scopeset mapping ."""

    def get_scopeset_names(self) -> list[str]:
        """
        Get the names of scopesets which have been set by model (not defaults).

        :return: set of names, for example ['gwp', 'odp']
        """
        result = []
        for f in self.__fields_set__:
            if f in ("ext",):
                continue
            field = self.__fields__.get(f)
            # field can be explicitly specified, or can be an unknown impact covered by extra='allow'
            result.append(field.alias if field and field.alias else f)

        return result

    def get_scopeset_by_name(self, name: str) -> ScopeSet | None:
        """
        Get scopeset by name.

        :param name: The name of the scopeset.
        :return: A scopeset if found, None otherwise
        """
        # check known impacts first
        for f_name, f in self.__fields__.items():
            if f.alias == name:
                return getattr(self, f_name)
            if f_name == name:
                return getattr(self, f_name)
        # probably unknown impact, coming from 'extra' fields
        return getattr(self, name, None)

    @pyd.root_validator(skip_on_failure=True)
    def _extra_scopeset_validator(cls, values: dict[str, Any]) -> dict[str, Any]:
        for f in values:
            # only interested in validating the extra fields
            if f in cls.__fields__:
                continue

            # extra impact of an unknown type - engage validation of ScopeSet
            extra_scopeset = values.get(f)
            match extra_scopeset:
                case ScopeSet():
                    continue
                case dict():
                    values[f] = ScopeSet(**extra_scopeset)
                case _:
                    raise ValueError(f"{f} must be a ScopeSet schema")

        return values


class ScopeSetGwp(ScopeSet):
    """ScopeSet measured in kgCO2e."""

    allowed_units = "kgCO2e"


class ScopeSetOdp(ScopeSet):
    """ScopeSet measured in kgCFC11e."""

    allowed_units = "kgCFC11e"


class ScopeSetAp(ScopeSet):
    """ScopeSet measured in kgSO2e."""

    allowed_units = ("kgSO2e", "molHe")


class ScopeSetEpNe(ScopeSet):
    """ScopeSet measured in kgNe."""

    allowed_units = "kgNe"


class ScopeSetPocp(ScopeSet):
    """ScopeSet measured in kgO3e."""

    allowed_units = ("kgO3e", "kgNMVOCe")


class ScopeSetEpFresh(ScopeSet):
    """ScopeSet measured in kgPO4e."""

    allowed_units = "kgPO4e"


class ScopeSetEpTerr(ScopeSet):
    """ScopeSet measured in molNe."""

    allowed_units = "molNe"


class ScopeSetIrp(ScopeSet):
    """ScopeSet measured in kilo Becquerel equivalent of u235."""

    allowed_units = "kBqU235e"


class ScopeSetCTUh(ScopeSet):
    """ScopeSet measured in CTUh."""

    allowed_units = "CTUh"


class ScopeSetM3Aware(ScopeSet):
    """ScopeSet measured in m3AWARE Water consumption by AWARE method."""

    allowed_units = "m3AWARE"


class ScopeSetCTUe(ScopeSet):
    """ScopeSet measured in CTUe."""

    allowed_units = "CTUe"


class ScopeSetDiseaseIncidence(ScopeSet):
    """ScopeSet measuring disease incidence measured in AnnualPerCapita (cases)."""

    allowed_units = "AnnualPerCapita"


class ScopeSetMass(ScopeSet):
    """ScopeSet measuring mass in kg."""

    allowed_units = "kg"


class ScopeSetVolume(ScopeSet):
    """ScopeSet measuring mass in kg."""

    allowed_units = "m3"


class ScopeSetMassOrVolume(ScopeSet):
    """ScopeSet measuring mass in kg OR volume in m3, example: radioactive waste."""

    allowed_units = ("kg", "m3")


class ScopeSetEnergy(ScopeSet):
    """ScopeSet measuring mass in kg."""

    allowed_units = "MJ"


class ImpactSet(ScopesetByNameBase):
    """A set of impacts, such as GWP, ODP, AP, EP, POCP, EP-marine, EP-terrestrial, EP-freshwater, etc."""

    gwp: ScopeSetGwp | None = pyd.Field(
        default=None,
        description="GWP100, calculated per IPCC guidelines.  If any CO2 removals are "
        "part of this figure, the gwp-fossil, gwp-bioganic, gwp-luluc, an "
        "gwp-nonCO2 fields are required, as is "
        "kg_C_biogenic_per_declared_unit.",
    )
    odp: ScopeSetOdp | None = pyd.Field(default=None, description="Ozone Depletion Potential")
    ap: ScopeSetAp | None = pyd.Field(default=None, description="Acidification Potential")
    ep: ScopeSetEpNe | None = pyd.Field(
        default=None, description="Eutrophication Potential in Marine Ecosystems. Has the same meaning as ep-marine."
    )
    pocp: ScopeSetPocp | None = pyd.Field(default=None, description="Photochemical Smog (Ozone) creation potential")
    ep_marine: ScopeSetEpNe | None = pyd.Field(
        alias="ep-marine", default=None, description="Has the same meaning as 'ep'"
    )
    ep_fresh: ScopeSetEpFresh | None = pyd.Field(
        alias="ep-fresh", default=None, description="Eutrophication Potential in Freshwater Ecosystems"
    )
    ep_terr: ScopeSetEpTerr | None = pyd.Field(
        alias="ep-terr", default=None, description="Eutrophication Potential in Terrestrial Ecosystems"
    )
    gwp_biogenic: ScopeSetGwp | None = pyd.Field(
        alias="gwp-biogenic",
        default=None,
        description="Net GWP from removals of atmospheric CO2 into biomass and  emissions of CO2 from biomass sources. "
        "To be counted as negative, CO2 removals must be closely linked to production in "
        "operational control (i.e. no purchased offsets), time (within three years of production) and "
        "space (similar biome).  They must not have been sold, committed, or credited to any other "
        "product.  Harvesting from native forests is handled under GWP_luluc for EN15804.",
    )
    gwp_luluc: ScopeSetGwp | None = pyd.Field(
        alias="gwp-luluc",
        default=None,
        description="Climate change effects related to land use and land use change, for example biogenic carbon "
        "exchanges resulting from deforestation or other soil activities (including soil carbon "
        "emissions). All related emissions for native forests are included under this category. "
        "Uptake for native forests is set to 0 kgCO2 for EN15804.",
    )
    gwp_nonCO2: ScopeSetGwp | None = pyd.Field(
        alias="gwp-nonCO2",
        default=None,
        description="GWP from non-CO2, non-fossil sources, such as livestock-sourced CH4 and agricultural N2O.",
    )
    gwp_fossil: ScopeSetGwp | None = pyd.Field(
        alias="gwp-fossil",
        default=None,
        description="Climate change effects due to greenhouse gas emissions originating from the oxidation or "
        "reduction of fossil fuels or materials containing fossil carbon. [Source: EN15804]",
    )
    WDP: ScopeSetM3Aware | None = pyd.Field(
        default=None,
        description="Deprivation-weighted water consumption, calculated by the AWARE method "
        "(https://wulca-waterlca.org/aware/what-is-aware)",
    )
    PM: ScopeSetDiseaseIncidence | None = pyd.Field(
        default=None,
        description="Potential incidence of disease due to particulate matter emissions.",
    )
    IRP: ScopeSetIrp | None = pyd.Field(
        default=None,
        description="Potential ionizing radiation effect on human health, relative to U235.",
    )
    ETP_fw: ScopeSetCTUe | None = pyd.Field(
        alias="ETP-fw",
        default=None,
        description="Ecotoxicity in freshwater, in potential Comparative Toxic Unit for ecosystems.",
    )
    HTP_c: ScopeSetCTUh | None = pyd.Field(
        alias="HTP-c",
        default=None,
        description="Human toxicity, cancer effects in potential Comparative Toxic Units for humans.",
    )
    HTP_nc: ScopeSetCTUh | None = pyd.Field(
        alias="HTP-nc",
        default=None,
        description="Human toxicity, noncancer effects in potential Comparative Toxic Units for humans.",
    )
    SQP: ScopeSet | None = pyd.Field(
        default=None,
        description="Land use related impacts / Soil quality, in potential soil quality parameters.",
    )


class LCIAMethod(StrEnum):
    """A list of available LCA methods."""

    UNKNOWN = "Unknown LCIA"
    TRACI_2_1 = "TRACI 2.1"
    TRACI_2_0 = "TRACI 2.0"
    TRACI_1_0 = "TRACI 1.0"
    IPCC_AR5 = "IPCC AR5"
    EF_3_0 = "EF 3.0"
    EF_3_1 = "EF 3.1"
    EF_2_0 = "EF 2.0"
    EN_15978_2011 = "EN 15978:2011"
    USETOX_2_12 = "USEtox 2.12"
    CML_2016 = "CML 2016"
    CML_2012 = "CML 2012"
    CML_2007 = "CML 2007"
    CML_2001 = "CML 2001"
    CML_1992 = "CML 1992"
    RECIPE_2016 = "ReCiPe 2016"
    RECIPE_2008 = "ReCiPe 2008"
    LIME2 = "LIME2"

    @classmethod
    def is_method_supported(cls, method_name: str | None) -> bool:
        """Return True if the method is supported, False otherwise."""
        if method_name is None:
            return False
        try:
            cls(method_name)
        except ValueError:
            return False
        return True

    @classmethod
    def get_by_name(cls, d_name: str | None) -> "LCIAMethod":
        """Return the LCIAMethod enum value for the given name, or UNKNOWN if not found."""
        if d_name is None:
            return cls.UNKNOWN
        try:
            return cls(d_name)
        except ValueError:
            return cls.UNKNOWN


class Impacts(pyd.BaseModel):
    """List of environmental impacts, compiled per one of the standard Impact Assessment methods."""

    class Config:
        # pydantic schema generator gets lost in this structure, so we need to re-establish it manually for openapi
        schema_extra = {
            "properties": {
                str(lm): {"description": str(lm), "allOf": [{"$ref": "#/components/schemas/ImpactSet"}]}
                for lm in LCIAMethod
            },
            "additionalProperties": None,
        }

    __root__: dict[LCIAMethod, ImpactSet]

    def set_unknown_lcia(self, impact_set: ImpactSet):
        """Set the impact set as an unknown LCIA method."""
        self.__root__[LCIAMethod.UNKNOWN] = impact_set

    def set_impact_set(self, lcia_method: LCIAMethod | str | None, impact_set: ImpactSet):
        """
        Set the impact set for the given LCIA method.

        If the LCIA method is None, set it as an unknown LCIA method.
        """
        if lcia_method is None:
            self.set_unknown_lcia(impact_set)
        else:
            if isinstance(lcia_method, str):
                lcia_method = LCIAMethod.get_by_name(lcia_method)
            self.__root__[lcia_method] = impact_set

    def replace_lcia_method(self, lcia_method: LCIAMethod, new_lcia_method: LCIAMethod) -> None:
        """
        Replace the LCIA method.

        If the there is no impact set for the given LCIA method, do nothing.
        """
        impact_set = self.get_impact_set(lcia_method)
        if impact_set is None:
            return None
        self.set_impact_set(new_lcia_method, impact_set)
        del self.__root__[lcia_method]

    def get_impact_set(
        self, lcia_method: LCIAMethod | str | None, default_val: ImpactSet | None = None
    ) -> ImpactSet | None:
        """Return the impact set for the given LCIA method."""
        if lcia_method is None:
            return self.__root__.get(LCIAMethod.UNKNOWN, default_val)
        if isinstance(lcia_method, str):
            lcia_method = LCIAMethod.get_by_name(lcia_method)
        return self.__root__.get(lcia_method, default_val)

    def available_methods(self) -> set[LCIAMethod]:
        """Return a list of available LCIA methods."""
        return set(self.__root__.keys())

    def as_dict(self) -> dict[LCIAMethod, ImpactSet]:
        """Return the impacts as a dictionary."""
        return self.__root__


class ResourceUseSet(ScopesetByNameBase):
    """A set of resource use indicators, such as RPRec, RPRm, etc."""

    RPRec: ScopeSetEnergy | None = pyd.Field(
        description="Renewable primary resources used as energy carrier (fuel). "
        "First use bio-based materials used as an energy source. Hydropower, solar and wind power used "
        "in the technosphere are also included in this indicator",
        default=None,
    )
    RPRm: ScopeSetEnergy | None = pyd.Field(
        description="Renewable primary resources with energy content used as material. "
        "First use biobased materials used as materials (e.g. wood, hemp, etc.).",
        default=None,
    )
    rpre: ScopeSetEnergy | None = pyd.Field(
        description="Renewable primary energy resources as energy",
        default=None,
    )
    nrpre: ScopeSetEnergy | None = pyd.Field(
        description="Non-renewable primary resources as energy (fuel)",
        default=None,
    )
    nrprm: ScopeSetEnergy | None = pyd.Field(
        description="Non-renewable primary resources as material",
        default=None,
    )
    fw: ScopeSetVolume | None = pyd.Field(
        description="Use of net fresh water",
        default=None,
    )
    sm: ScopeSetMass | None = pyd.Field(
        description="Use of secondary materials",
        default=None,
    )
    rsf: ScopeSetEnergy | None = pyd.Field(
        description="Use of renewable secondary materials",
        default=None,
    )
    nrsf: ScopeSetEnergy | None = pyd.Field(
        description="Use of non-renewable secondary fuels",
        default=None,
    )
    re: ScopeSetEnergy | None = pyd.Field(
        description="Renewable energy resources",
        default=None,
    )
    pere: ScopeSetEnergy | None = pyd.Field(
        description="Use of renewable primary energy excluding renewable primary energy resources used as raw materials",
        default=None,
    )
    perm: ScopeSetEnergy | None = pyd.Field(
        description="Use of renewable primary energy resources used as raw materials",
        default=None,
    )
    pert: ScopeSetEnergy | None = pyd.Field(
        description="Total use of renewable primary energy resources",
        default=None,
    )
    penre: ScopeSetEnergy | None = pyd.Field(
        description="Use of non-renewable primary energy excluding "
        "non-renewable primary energy resources used as raw materials",
        default=None,
    )
    penrm: ScopeSetEnergy | None = pyd.Field(
        description="Use of non-renewable primary energy resources used as raw materials",
        default=None,
    )
    penrt: ScopeSetEnergy | None = pyd.Field(
        description="Total use of non-renewable primary energy resources",
        default=None,
    )


class OutputFlowSet(ScopesetByNameBase):
    """A set of output flows, such as waste, emissions, etc."""

    twd: ScopeSetMass | None = pyd.Field(
        description="Total waste disposed",
        default=None,
    )
    hwd: ScopeSetMass | None = pyd.Field(
        description="Hazardous waste disposed",
        default=None,
    )
    nhwd: ScopeSetMass | None = pyd.Field(
        description="Non-hazardous waste disposed",
        default=None,
    )
    rwd: ScopeSetMass | None = pyd.Field(
        description="Radioactive waste disposed",
        default=None,
    )
    hlrw: ScopeSetMassOrVolume | None = pyd.Field(
        description="High level radioactive waste disposed",
        default=None,
    )
    illrw: ScopeSetMassOrVolume | None = pyd.Field(
        description="Intermediate level radioactive waste disposed",
        default=None,
    )
    cru: ScopeSetMass | None = pyd.Field(
        description="Components for re-use",
        default=None,
    )
    mr: ScopeSetMass | None = pyd.Field(
        description="Recycled materials",
        default=None,
    )
    mfr: ScopeSetMass | None = pyd.Field(
        description="Materials for recycling",
        default=None,
    )
    mer: ScopeSetMass | None = pyd.Field(
        description="Materials for energy recovery",
        default=None,
    )
    ee: ScopeSetEnergy | None = pyd.Field(
        description="Exported energy",
        default=None,
    )
    eh: ScopeSetEnergy | None = pyd.Field(
        description="Exported heat",
        default=None,
    )


class WithLciaMixin(BaseOpenEpdSchema):
    """Mixin for LCIA data."""

    impacts: Impacts | None = pyd.Field(
        description="List of environmental impacts, compiled per one of the standard Impact Assessment methods",
        example={"TRACI 2.1": {"gwp": {"A1A2A3": {"mean": 22.4, "unit": "kgCO2e"}}}},
    )
    resource_uses: ResourceUseSet | None = pyd.Field(
        description="Set of Resource Use Indicators, over various LCA scopes",
        example={"RPRe": {"A1A2A3": {"mean": 12, "unit": "MJ", "rsd": 0.12}}},
    )
    output_flows: OutputFlowSet | None = pyd.Field(
        description="Set of Waste and Output Flow indicators which describe the waste categories "
        "and other material output flows derived from the LCI.",
        example={"hwd": {"A1A2A3": {"mean": 2300, "unit": "kg", "rsd": 0.22}}},
    )



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from enum import StrEnum

from openepd.compat.pydantic import pyd
from openepd.model.common import OpenEPDUnit
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.validation.numbers import RatioFloat
from openepd.model.validation.quantity import LengthMmStr, TemperatureCStr, validate_quantity_unit_factory


class AsphaltMixType(StrEnum):
    """Asphalt mix type enum."""

    HMA = "HMA"
    WMA = "WMA"


class AsphaltGradation(StrEnum):
    """Asphalt gradation enum."""

    Dense_graded = "Dense-graded"
    Open_graded = "Open-graded"
    Gap_graded = "Gap-graded"
    Permeable = "Permeable"
    Porous = "Porous"
    Other = "Other"


class AsphaltV1(BaseOpenEpdHierarchicalSpec):
    """Asphalt spec."""

    _EXT_VERSION = "1.1"

    asphalt_aggregate_size_max: LengthMmStr | None = pyd.Field(
        default=None, example="5mm", description="Max aggregate size"
    )

    asphalt_rap: RatioFloat | None = pyd.Field(
        default=None, description="Percent of mixture that has been replaced by recycled " "asphalt pavement (RAP)."
    )
    asphalt_ras: RatioFloat | None = pyd.Field(
        default=None, description="Percent of mixture that has been replaced by recycled " "asphalt shingles (RAS)."
    )
    asphalt_ground_tire_rubber: RatioFloat | None = pyd.Field(
        default=None, description="Percent of mixture that has been replaced " "by ground tire rubber (GTR)."
    )

    asphalt_max_temperature: TemperatureCStr | None = pyd.Field(
        default=None,
        description="The upper threshold temperature to which an asphalt "
        "binder can be heated preventing the asphalt mixture "
        "from rutting",
    )
    asphalt_min_temperature: TemperatureCStr | None = pyd.Field(
        default=None,
        description="The lower threshold temperature for an asphalt "
        "binder to prevent thermal cracking of the asphalt"
        " mixture.",
    )

    asphalt_mix_type: AsphaltMixType | None = pyd.Field(default=None, description="Asphalt mix type")
    asphalt_gradation: AsphaltGradation | None = pyd.Field(default=None, description="Asphalt gradation")

    asphalt_sbr: bool | None = pyd.Field(default=None, description="Styrene-butadiene rubber (SBR)")
    asphalt_sbs: bool | None = pyd.Field(default=None, description="Styrene-butadiene-styrene (SBS)")
    asphalt_ppa: bool | None = pyd.Field(default=None, description="Polyphosphoric acid (PPA)")
    asphalt_gtr: bool | None = pyd.Field(default=None, description="Ground tire rubber (GTR)")
    asphalt_pmb: bool | None = pyd.Field(default=None, description="Polymer modified bitumen (PMB)")

    _aggregate_size_max_validator = pyd.validator("asphalt_aggregate_size_max", allow_reuse=True)(
        validate_quantity_unit_factory(OpenEPDUnit.m)
    )



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Annotated, Literal

from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec, CodegenSpec
from openepd.model.specs.concrete import Cementitious, ConcreteTypicalApplication
from openepd.model.specs.enums import AciExposureClass, CsaExposureClass, EnExposureClass
from openepd.model.validation.enum import exclusive_groups_validator_factory
from openepd.model.validation.numbers import RatioFloat
from openepd.model.validation.quantity import (
    LengthInchStr,
    LengthMmStr,
    MassKgStr,
    PressureMPaStr,
    validate_quantity_unit_factory,
)


class CementGroutV1(BaseOpenEpdHierarchicalSpec):
    """
    Cement grout performance specification.

    Cement grouting is a slurry that is placed as a flowable liquid. It is an effective material
    for filling and strengthening granular soils, voids in rocks, foundation underpinnings, and
    other underground voids. Also called structural grout, these materials typically impart
    significant compressive strength to the system.
    """

    _EXT_VERSION = "1.0"


class ConcretePavingV1(BaseOpenEpdHierarchicalSpec):
    """Concrete paving."""

    _EXT_VERSION = "1.0"

    # Own fields:
    flexion_strength: PressureMPaStr | None = pyd.Field(
        default=None, description="Concrete flexural strength.", example="30 MPa"
    )

    _concrete_flexion_strength_is_quantity_validator = pyd.validator("flexion_strength", allow_reuse=True)(
        validate_quantity_unit_factory("MPa")
    )


class FlowableFillV1(BaseOpenEpdHierarchicalSpec):
    """
    Flowable fill performance specification.

    Flowable fill is a slurry that is placed as a flowable liquid (high slump) and sets with no
    compaction. It is often used in tight or restricted access areas where placing and compacting
    fill is difficult. Applications include filling large voids such as abandoned underground storage
    tanks, basements, tunnels, mines, and sewers. It can also be used as paving sub-base, bridge
    abutment, and retaining wall backfill.

    Also called Controlled Density Fill (CDF) or Controlled Low Strength Materials (CLSMs). These materials typically
    have compressive strengths under 1200 psi.
    """

    _EXT_VERSION = "1.0"


class OilPatchV1(BaseOpenEpdHierarchicalSpec):
    """
    Petroleum Industry Cement Slurry.

    Concretes for use in creation, maintenance, and decommissioning of petroleum extraction wells and similar
    applications. Includes foamed cement; often called cement in the drilling industry. Differs from
    flowable fill and grout in that it contains no sand or other aggregates.
    """

    _EXT_VERSION = "1.0"


class ReadyMixV1(BaseOpenEpdHierarchicalSpec):
    """Concretes to be mixed and then poured on-site."""

    _EXT_VERSION = "1.0"


class ShotcreteV1(BaseOpenEpdHierarchicalSpec):
    """Concretes sprayed on a target."""

    _EXT_VERSION = "1.0"


class ConcreteV1(BaseOpenEpdHierarchicalSpec):
    """
    Concrete.

    A composite material composed of fine and coarse aggregate bonded together with a fluid cement (cement paste) that
    hardens over time.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    lightweight: bool | None = pyd.Field(default=None, description="Product is lightweight", example=True)

    strength_28d: PressureMPaStr | None = pyd.Field(
        default=None, description="Concrete strength after 28 days", example="1 MPa"
    )
    strength_other: PressureMPaStr | None = pyd.Field(
        default=None,
        description="A strength spec which is to be reached later other 28 days (e.g. 42d)",
        example="30 MPa",
    )
    strength_other_d: Annotated[
        Literal[3, 7, 14, 42, 56, 72, 96, 120] | None,
        CodegenSpec(override_type=Literal[3, 7, 14, 42, 56, 72, 96, 120]),
    ] = pyd.Field(default=None, description="Test Day for strength_other", example=42)

    slump: LengthInchStr | None = pyd.Field(default=None, description="", example="2 in")
    min_slump: LengthInchStr | None = pyd.Field(default=None, description="Minimum test slump", example="2 in")
    max_slump: LengthInchStr | None = pyd.Field(default=None, description="", example="2 in")

    min_pipeline_size: LengthMmStr | None = pyd.Field(
        default=None, description="Minimum pipeline size", example="200 mm"
    )
    w_c_ratio: RatioFloat | None = pyd.Field(
        default=None, description="Ratio of water to cement", example=0.5, ge=0, le=1
    )
    air_entrain: bool | None = pyd.Field(default=None, description="Air Entrainment", example=True)
    co2_entrain: bool | None = pyd.Field(default=None, description="CO2 Curing", example=True)
    self_consolidating: bool | None = pyd.Field(default=None, description="Self Compacting", example=True)
    white_cement: bool | None = pyd.Field(default=None, description="White Cement", example=True)
    plc: bool | None = pyd.Field(default=None, description="Portland Limestone Cement", example=True)
    finishable: bool | None = pyd.Field(default=None, description="Finishable", example=True)
    fiber_reinforced: bool | None = pyd.Field(default=None, description="fiber_reinforced", example=True)

    cementitious: Cementitious | None = pyd.Field(
        default=None, description="List of cementitious materials, and proportion by mass"
    )

    aggregate_size_max: LengthMmStr | None = pyd.Field(
        default=None,
        description="The smallest sieve size for which the entire amount of aggregate is able to pass. "
        "Parameter describes diameter of aggregate",
        example="8 mm",
    )
    cement_content: MassKgStr | None = pyd.Field(default=None, example="1 kg")

    aci_exposure_classes: list[AciExposureClass] | None = pyd.Field(
        default=None, description="List of ACI exposure classes", example=["aci.F0"]
    )
    csa_exposure_classes: list[CsaExposureClass] | None = pyd.Field(
        default=None, description="List of CSA exposure classes", example=["csa.C-2"]
    )
    en_exposure_classes: list[EnExposureClass] | None = pyd.Field(
        default=None, description="List of EN exposure classes", example=["en206.X0"]
    )
    typical_application: ConcreteTypicalApplication | None = pyd.Field(default=None, description="Typical Application")

    # Nested specs:
    CementGrout: CementGroutV1 | None = None
    ConcretePaving: ConcretePavingV1 | None = None
    FlowableFill: FlowableFillV1 | None = None
    OilPatch: OilPatchV1 | None = None
    ReadyMix: ReadyMixV1 | None = None
    Shotcrete: ShotcreteV1 | None = None

    _aci_exposure_classes_exclusive_groups_validator = pyd.validator("aci_exposure_classes", allow_reuse=True)(
        exclusive_groups_validator_factory(AciExposureClass)
    )
    _en_exposure_classes_exclusive_groups_validator = pyd.validator("en_exposure_classes", allow_reuse=True)(
        exclusive_groups_validator_factory(EnExposureClass)
    )
    _csa_exposure_classes_exclusive_groups_validator = pyd.validator("csa_exposure_classes", allow_reuse=True)(
        exclusive_groups_validator_factory(CsaExposureClass)
    )



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "ConveyorsRangeV1",
    "MaterialHandlingRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec


class ConveyorsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Machinery and tools designed to move materials within a facility, both manually and automatically.

    This includes various types of conveyors such as belt, roller, and overhead conveyors.

    Range version.
    """

    _EXT_VERSION = "1.0"


class MaterialHandlingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Material handling.

    Equipment and supplies for moving, storing, administering, and protecting materials during the handling process.

    Range version.
    """

    _EXT_VERSION = "1.0"

    Conveyors: ConveyorsRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import AluminiumAlloy


class AluminiumBilletsV1(BaseOpenEpdHierarchicalSpec):
    """Cast Aluminium ingots or billets for use in manufacturing more finished products."""

    _EXT_VERSION = "1.0"


class AluminiumExtrusionsV1(BaseOpenEpdHierarchicalSpec):
    """Extruded aluminum products used in construction.

    Includes range of finish options including mill finish, painted, and anodized.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    thermally_improved: bool | None = pyd.Field(default=None, description="", example=True)


class AluminiumSheetGoodsV1(BaseOpenEpdHierarchicalSpec):
    """Rolled and/or Stamped Aluminium coil or sheets, often used in flashing, trim, panels, and deck."""

    _EXT_VERSION = "1.0"


class AluminiumSuspensionAssemblyV1(BaseOpenEpdHierarchicalSpec):
    """Aluminum suspension assemblies for acoustical ceiling systems."""

    _EXT_VERSION = "1.0"


class AluminiumV1(BaseOpenEpdHierarchicalSpec):
    """Broad category for construction materials made primarily from Aluminium and its alloys."""

    _EXT_VERSION = "1.0"

    # Own fields:
    alloy: AluminiumAlloy | None = pyd.Field(default=None, description="", example=str(AluminiumAlloy.ALLOY_1XXX))
    anodized: bool | None = pyd.Field(default=None, description="", example=True)
    painted: bool | None = pyd.Field(default=None, description="", example=True)

    # Nested specs:
    AluminiumBillets: AluminiumBilletsV1 | None = None
    AluminiumExtrusions: AluminiumExtrusionsV1 | None = None
    AluminiumSheetGoods: AluminiumSheetGoodsV1 | None = None
    AluminiumSuspensionAssembly: AluminiumSuspensionAssemblyV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
try:
    from pydantic import functional_validators  # type: ignore
    from pydantic import v1 as pyd  # type: ignore
    from pydantic.v1 import generics as pyd_generics  # type: ignore

except ImportError:
    import pydantic as pyd  # type: ignore[no-redef]
    from pydantic import generics as pyd_generics  # type: ignore[no-redef]

    from . import compat_functional_validators as functional_validators  # type: ignore[no-redef]


pydantic = pyd

__all__ = ["pyd", "pydantic", "pyd_generics", "functional_validators"]



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import (
    IntumescentFireproofingMaterialType,
    SprayFireproofingDensity,
    SprayFireproofingMaterialType,
)
from openepd.model.validation.quantity import LengthMmStr


class IntumescentFireproofingV1(BaseOpenEpdHierarchicalSpec):
    """
    Fireproofing material applied to structural materials, which swells as a result of heat exposure.

    As a result it increases in volume and decreasing in density.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    material_type: IntumescentFireproofingMaterialType | None = pyd.Field(default=None, description="", example="Epoxy")


class SprayFireproofingV1(BaseOpenEpdHierarchicalSpec):
    """
    Spray fireproofing.

    A passive fire protection system that reduces the rate of temperature increase in concrete or steel during a fire.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    material_type: SprayFireproofingMaterialType | None = pyd.Field(
        default=None, description="", example="Gypsum-based"
    )
    density: SprayFireproofingDensity | None = pyd.Field(default=None, description="", example="Standard")


class AppliedFireproofingV1(BaseOpenEpdHierarchicalSpec):
    """
    Fireproofing material applied to structural materials.

    Materials include: cement aggregate, cementitious, magnesium-oxychloride, intumescent, magnesium cement, mineral
    fiber, and mineral fiber fireproofing products.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: LengthMmStr | None = pyd.Field(default=None, description="", example="10 mm")

    # Nested specs:
    IntumescentFireproofing: IntumescentFireproofingV1 | None = None
    SprayFireproofing: SprayFireproofingV1 | None = None


class FirestoppingV1(BaseOpenEpdHierarchicalSpec):
    """
    Seals and protects openings and joints in fire rate assemblies.

    Typically sealants, sprays, and caulks.
    """

    _EXT_VERSION = "1.0"


class FireAndSmokeProtectionV1(BaseOpenEpdHierarchicalSpec):
    """
    Fire and smoke protection.

    General category of materials whose function is to provide protection of materials, spaces, and occupants from
    fire and smoke damage.
    """

    _EXT_VERSION = "1.0"

    # Nested specs:
    AppliedFireproofing: AppliedFireproofingV1 | None = None
    Firestopping: FirestoppingV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.api.base_sync_client import BaseApiMethodGroup
from openepd.api.category.dto import CategoryTreeResponse
from openepd.model.category import Category


class CategoryApi(BaseApiMethodGroup):
    """API methods for reading categories."""

    def get_tree_raw(self) -> CategoryTreeResponse:
        """
        Get categories tree.

        :return: categories tree wrapped in OpenEpdApiResponse
        """
        response = self._client.do_request("get", "/v2/categories/tree")
        return CategoryTreeResponse.parse_raw(response.content)

    def get_tree(self) -> Category:
        """
        Get categories tree.

        :return: categories tree
        """
        response = self.get_tree_raw()
        return response.payload



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import FireProtectionPipingMaterial, PipingAnsiSchedule, PlumbingPipingMaterial
from openepd.model.validation.quantity import LengthMmStr, MassPerLengthStr


class BathtubsV1(BaseOpenEpdHierarchicalSpec):
    """Bathtubs."""

    _EXT_VERSION = "1.0"


class FaucetsV1(BaseOpenEpdHierarchicalSpec):
    """Faucets."""

    _EXT_VERSION = "1.0"


class OtherPlumbingFixturesV1(BaseOpenEpdHierarchicalSpec):
    """Other plumbing fixtures."""

    _EXT_VERSION = "1.0"


class WaterClosetsV1(BaseOpenEpdHierarchicalSpec):
    """Water Closets."""

    _EXT_VERSION = "1.0"


class FireProtectionPipingV1(BaseOpenEpdHierarchicalSpec):
    """System of pipes used to supply fire-suppression fluids to homes and/or businesses."""

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: LengthMmStr | None = pyd.Field(default=None, description="", example="6 mm")
    piping_diameter: LengthMmStr | None = pyd.Field(default=None, description="", example="120 mm")
    mass_per_unit_length: MassPerLengthStr | None = pyd.Field(default=None, description="", example="1 kg / m")
    piping_ansi_schedule: PipingAnsiSchedule | None = pyd.Field(default=None, description="", example="5")
    fire_protection_piping_material: FireProtectionPipingMaterial | None = pyd.Field(
        default=None, description="", example="PVC"
    )


class FireSprinklersV1(BaseOpenEpdHierarchicalSpec):
    """Fire sprinklers."""

    _EXT_VERSION = "1.0"


class StorageTanksV1(BaseOpenEpdHierarchicalSpec):
    """Storage tanks."""

    _EXT_VERSION = "1.0"


class WaterHeatersV1(BaseOpenEpdHierarchicalSpec):
    """Water heaters."""

    _EXT_VERSION = "1.0"


class PlumbingFixturesV1(BaseOpenEpdHierarchicalSpec):
    """Residential and commercial water closets, urinals, lavatories, sinks, bathtubs, showers, faucets, etc."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    Bathtubs: BathtubsV1 | None = None
    Faucets: FaucetsV1 | None = None
    OtherPlumbingFixtures: OtherPlumbingFixturesV1 | None = None
    WaterClosets: WaterClosetsV1 | None = None


class FireSuppressionV1(BaseOpenEpdHierarchicalSpec):
    """Systems used to extinguish, control, or prevent fires."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    FireProtectionPiping: FireProtectionPipingV1 | None = None
    FireSprinklers: FireSprinklersV1 | None = None


class PipingV1(BaseOpenEpdHierarchicalSpec):
    """
    Piping.

    System of pipes used to provide water and fuel, remove wastewater, allow venting of gases, or supply
    fire-suppression fluids to homes, businesses, or other facilities.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: LengthMmStr | None = pyd.Field(default=None, description="", example="6 mm")
    piping_diameter: LengthMmStr | None = pyd.Field(default=None, description="", example="120 mm")
    mass_per_unit_length: MassPerLengthStr | None = pyd.Field(default=None, description="", example="1 kg / m")
    piping_ansi_schedule: PipingAnsiSchedule | None = pyd.Field(default=None, description="", example="5")
    plumbing_piping_material: PlumbingPipingMaterial | None = pyd.Field(default=None, description="", example="PVC")


class PlumbingEquipmentV1(BaseOpenEpdHierarchicalSpec):
    """Water softeners, filtration equipment, water heaters, and other plumbing equipment."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    StorageTanks: StorageTanksV1 | None = None
    WaterHeaters: WaterHeatersV1 | None = None


class PlumbingV1(BaseOpenEpdHierarchicalSpec):
    """Residential and commercial plumbing equipment and fixtures."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    PlumbingFixtures: PlumbingFixturesV1 | None = None
    FireSuppression: FireSuppressionV1 | None = None
    Piping: PipingV1 | None = None
    PlumbingEquipment: PlumbingEquipmentV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec, BaseOpenEpdSpec
from openepd.model.validation.numbers import RatioFloat
from openepd.model.validation.quantity import PressureMPaStr


class StructuralPrecastElementType(BaseOpenEpdSpec):
    """Precast element types for Structural Precast Concrete."""

    wall: bool | None = pyd.Field(
        description="Precast solid wall elements, including thin shell",
        example=True,
        default=None,
    )
    solid_slab: bool | None = pyd.Field(
        description="Precast slabs used for floor or roof applications",
        example=True,
        default=None,
    )
    hollowcore: bool | None = pyd.Field(
        description="Precast slabs with tubular voids, typically used flooring applications",
        example=True,
        default=None,
    )
    beam: bool | None = pyd.Field(
        description="Precast structural beams for carrying or transferring loads; includes girders",
        example=True,
        default=None,
    )
    column: bool | None = pyd.Field(
        description="Precast structural columns for carrying or transferring loads",
        example=True,
        default=None,
    )
    stairs: bool | None = pyd.Field(
        description="Staircases made of precast concrete components",
        example=True,
        default=None,
    )
    balcony: bool | None = pyd.Field(
        description="Balcony slabs made from precast concrete",
        example=True,
        default=None,
    )
    pile: bool | None = pyd.Field(
        description="Precast concrete structural foundation elements used to support offshore structures such as bridges, oil rigs, and floating airports",
        example=True,
        default=None,
    )


class UtilityPrecastElementType(BaseOpenEpdSpec):
    """Precast element types for Utility Underground Precast Concrete."""

    manhole: bool | None = pyd.Field(
        description="Precast barrel-shaped chambers used for wastewater management and access management",
        example=True,
        default=None,
    )
    retaining_wall: bool | None = pyd.Field(
        description="Precast concrete structures for retaining soil",
        example=True,
        default=None,
    )
    box_culvert: bool | None = pyd.Field(
        description="A precast concrete structure commonly used to channel water, primarily as part of a drainage system",
        example=True,
        default=None,
    )


class ArchitecturalPrecastElementType(BaseOpenEpdSpec):
    """Precast element types for Architectural Precast Concrete."""

    wall: bool | None = pyd.Field(
        description="Precast solid wall elements, including thin shell",
        example=True,
        default=None,
    )


class CivilPrecastElementType(BaseOpenEpdSpec):
    """Precast element types for Civil Precast Concrete."""

    beam: bool | None = pyd.Field(
        description="Precast structural beams for carrying or transferring loads; includes girders",
        example=True,
        default=None,
    )
    manhole: bool | None = pyd.Field(
        description="Precast barrel-shaped chambers used for wastewater management and access management",
        example=True,
        default=None,
    )
    retaining_wall: bool | None = pyd.Field(
        description="Precast concrete structures for retaining soil",
        example=True,
        default=None,
    )
    rail_sleeper: bool | None = pyd.Field(
        description="Rectangular supports for the rails on railroad tracks, which transfer loads to the track ballast and subgrade and keep the rails spaced to the correct gauge. Also called railroad ties",
        example=True,
        default=None,
    )
    box_culvert: bool | None = pyd.Field(
        description="A precast concrete structure commonly used to channel water, primarily as part of a drainage system",
        example=True,
        default=None,
    )
    pile: bool | None = pyd.Field(
        description="Precast concrete structural foundation elements used to support offshore structures such as bridges, oil rigs, and floating airports",
        example=True,
        default=None,
    )
    road_barriers: bool | None = pyd.Field(
        description="Precast Vehicle and Traffic Barriers",
        example=True,
        default=None,
    )


class ArchitecturalPrecastV1(BaseOpenEpdHierarchicalSpec):
    """Precast concrete cladding used for architectural purposes."""

    _EXT_VERSION = "1.1"

    element_type: ArchitecturalPrecastElementType | None = pyd.Field(
        default=None, description="Precast element type used for architectural applications."
    )


class StructuralPrecastV1(BaseOpenEpdHierarchicalSpec):
    """Precast concrete used for structural purposes."""

    _EXT_VERSION = "1.1"

    element_type: StructuralPrecastElementType | None = pyd.Field(
        default=None, description="Precast element type used for structural applications."
    )


class UtilityUndergroundPrecastV1(BaseOpenEpdHierarchicalSpec):
    """Precast concrete for utility vaults, manhole, drain inlets. Excludes piping."""

    _EXT_VERSION = "1.1"

    element_type: UtilityPrecastElementType | None = pyd.Field(
        default=None, description="Precast element type used for utility underground applications."
    )


class CivilPrecastV1(BaseOpenEpdHierarchicalSpec):
    """Precast concrete used for civil engineering applications including bridges, highways, and railroads."""

    _EXT_VERSION = "1.0"

    element_type: CivilPrecastElementType | None = pyd.Field(
        default=None, description="Precast element type used as civil engineering components."
    )


class PrecastConcreteV1(BaseOpenEpdHierarchicalSpec):
    """General category for precast concrete components."""

    _EXT_VERSION = "1.1"

    # Own fields:
    lightweight: bool | None = pyd.Field(default=None, description="", example=True)
    concrete_compressive_strength_28d: PressureMPaStr | None = pyd.Field(default=None, description="", example="1 MPa")
    insulated: bool | None = pyd.Field(default=None, description="", example=True)
    gfrc: bool | None = pyd.Field(
        default=None,
        description="Glass Fiber Reinforced Concrete is fiber-reinforced concrete sometimes used in "
        "architectural panels",
        example=True,
    )
    steel_mass_percentage: RatioFloat | None = pyd.Field(default=None, description="", example=0.5, ge=0, le=1)

    # Nested specs:
    ArchitecturalPrecast: ArchitecturalPrecastV1 | None = None
    StructuralPrecast: StructuralPrecastV1 | None = None
    UtilityUndergroundPrecast: UtilityUndergroundPrecastV1 | None = None
    CivilPrecast: CivilPrecastV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from collections.abc import Iterable
from contextlib import contextmanager
from datetime import datetime, timedelta
import threading
from time import sleep
from typing import Callable, Generic, Iterator, cast

from requests import Response

from openepd.api.dto.common import DEFAULT_PAGE_SIZE, MetaCollectionDto, OpenEpdApiResponse
from openepd.api.dto.meta import PagingMeta, PagingMetaMixin
from openepd.model.base import TOpenEpdObject


class Throttler:
    """Throttle calls to a function to a certain rate."""

    def __init__(self, rate_per_sec: float) -> None:
        """
        Construct a throttler.

        :param rate_per_sec: number of calls to throttle per second
        """
        super().__init__()

        self.rate: float = rate_per_sec

        self.__count_lock = threading.Lock()
        self.__time_lock = threading.Lock()

        self.__count = 0
        self.__start = datetime.now()
        self.__time_limit = timedelta(seconds=1)

    @contextmanager
    def throttle(self):
        """Create context manager which throttles inside the context."""
        with self.__count_lock:
            count = self.__count

        with self.__time_lock:
            diff = datetime.now() - self.__start
            if diff < self.__time_limit and count >= self.rate:
                seconds = (self.__time_limit - diff).total_seconds()
                sleep(seconds)
                self.__start = datetime.now()
                self.__count = count - self.rate

        with self.__count_lock:
            self.__count += 1

        yield


class StreamingListResponse(Iterable[TOpenEpdObject], Generic[TOpenEpdObject]):
    """
    Iterator over a list of objects which could be from remote API in batches by given fetch function.

    This class could work only with OpenEpdApiResponse which:
    a) contains a list of objects in payload
    b) contains PagingMeta in meta

    Typical use case:

        stream: StreamingListResponse[Epd] # Assume we collected this from somewhere (e.g. from API Client)

        # Iterate over all EPDs (this will fetch the next page when needed, one single page is buffered)
        for epd in stream:
            print(epd)
            print("Current page: ", stream.current_page)

        # Get specific page
        page = stream.goto_page(4)

        # Get paging information
        stream.get_paging_meta()
        # or
        stream.get_get_total_pages(), stream.get_total_items()

        # Iterate starting from specific page
        for epd in stream.iterator(4):
            print(epd)
        # Alternative approach
        stream.goto_page(4)
        for epd in stream:
            print(epd)
    """

    def __init__(
        self,
        fetch_handler: Callable[[int, int], OpenEpdApiResponse[list[TOpenEpdObject], MetaCollectionDto]],
        auto_init: bool = True,
        page_size: int | None = None,
    ):
        self.__fetch_handler = fetch_handler
        self.__page_size = page_size or DEFAULT_PAGE_SIZE
        self.__current_page = 0
        self.__recent_response: OpenEpdApiResponse[list[TOpenEpdObject], MetaCollectionDto] | None = None
        if auto_init:
            self.goto_page(1)

    def goto_page(self, page_num: int, force_reload: bool = False) -> list[TOpenEpdObject]:
        """
        Go to the given page and return items as result.

        This will query remote server to retrieve given page, the state of iterator will be updated accordingly.
        All information from the internal buffer will be overridden.

        :param page_num: page number to retrieve
        :param force_reload: if True, force reload of the page even if it was already loaded
        :return: list of items on the page
        """
        if page_num <= 0:
            raise ValueError("Page number must be positive")
        if self.__current_page != page_num or force_reload:
            self.__recent_response = self.__fetch_handler(page_num, self.__page_size)
            self.__current_page = page_num
        if self.__recent_response is None:
            raise RuntimeError("Response is empty, this should not happen, check if fetch_handler is compatible")
        if self.__recent_response.payload is None:
            raise ValueError("Response does not contain payload")
        if not isinstance(self.__recent_response.payload, list):
            raise ValueError("Response does not contain a list")
        if self.__recent_response.meta is None:
            raise ValueError("Response does not contain meta")
        if not isinstance(self.__recent_response.meta, PagingMetaMixin):
            raise ValueError("Response does not contain paging meta")
        return self.__recent_response.payload

    def get_paging_meta(self) -> PagingMeta:
        """
        Get paging meta from the most recent response.

        If object is not initialized (no pages were retrieved yet), the first page will be fetched automatically.
        See also: get_meta()
        :return: paging meta
        """
        paging_meta = cast(PagingMetaMixin, self.get_meta()).paging
        if paging_meta is None:
            raise ValueError("Response does not contain paging meta")
        return paging_meta

    def get_meta(self) -> MetaCollectionDto:
        """
        Get meta from the most recent response.

        If object is not initialized (no pages were retrieved yet), the first page will be fetched automatically.
        :return: paging meta
        """
        self._ensure_initialized()
        return self.__recent_response.meta  # type: ignore

    @property
    def current_page(self) -> int:
        """Get current page number (numbering is 1-based)."""
        return self.__current_page

    def get_total_pages(self) -> int:
        """Get total number of pages."""
        return self.get_paging_meta().total_pages

    def get_total_count(self) -> int:
        """Get total number of items."""
        return self.get_paging_meta().total_count

    def has_next_page(self) -> bool:
        """Check if there is a next page."""
        return self.current_page < self.get_total_pages()

    def reset(self):
        """Reset iterator to the very beginning, cleanup internal buffers."""
        self.__current_page = 0
        self.__recent_response = None

    def __iter__(self) -> Iterator[TOpenEpdObject]:
        """
        Iterate over all items, when needed the new pages from server will be requested.

        Iteration from starts from the current page.
        """
        return self.iterator(self.current_page)

    def iterator(self, start_from_page: int = 1) -> Iterator[TOpenEpdObject]:
        """
        Iterate over all items, when needed the new pages from server will be requested.

        :param start_from_page: page number to start from (1-based)
        """
        if start_from_page <= 0:
            start_from_page = 1
        self.goto_page(start_from_page)
        while True:
            items = self.goto_page(self.current_page)
            for x in items:
                yield x
            if not self.has_next_page():
                return  # no more pages
            else:
                self.goto_page(self.current_page + 1)

    def __len__(self):
        return self.get_total_count()

    def _is_initialized(self) -> bool:
        """Check if object is initialized (at least one page was retrieved)."""
        return self.__recent_response is not None

    def _ensure_initialized(self):
        """Ensure that object is initialized, if not - request the first page."""
        if not self._is_initialized():
            self.goto_page(1)


def no_trailing_slash(val: str) -> str:
    """
    Remove all trailing slashes from the given string. Might be useful to normalize URLs.

    Non-string input is considered as error.
    """
    while val.endswith("/"):
        val = val[:-1]
    return val


def paging_meta_from_v1_api(r: Response) -> PagingMeta:
    """Return a PagingMeta dto from v1 API format with data in headers."""
    return PagingMeta(
        total_count=int(r.headers["X-Total-Count"]),
        total_pages=int(r.headers["X-Total-Pages"]),
        page_size=int(r.headers["X-Page-Size"]),
    )



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import TypeAlias

from openepd.api.dto.base import BaseOpenEpdApiModel
from openepd.api.dto.common import BaseMeta, OpenEpdApiResponse
from openepd.api.dto.meta import PagingMetaMixin, WarningMetaMixin
from openepd.api.dto.mf import MaterialFilterMetaMixin
from openepd.compat.pydantic import pyd
from openepd.model.common import Amount
from openepd.model.epd import Epd


class StatisticsDto(BaseOpenEpdApiModel):
    """
    DTO describes statistics of EPD.

    Statistics contains aggregated parameters such as percentiles distributions of GWP and other parameters.
    """

    # percentiles
    pct10_gwp: float = pyd.Field(
        description="10th percentile GWP for this statistics measured in kgCO2e per declared unit"
    )
    achievable_target: float = pyd.Field(
        description="Achievable target. 20th percentile of GWP measured in kgCO2e per declared unit", example=445.65
    )
    pct30_gwp: float = pyd.Field(
        description="30th percentile GWP for this statistics measured in kgCO2e per declared unit"
    )
    pct40_gwp: float = pyd.Field(
        description="40th percentile GWP for this statistics measured in kgCO2e per declared unit"
    )
    pct50_gwp: float = pyd.Field(
        description="50th percentile GWP for this statistics measured in kgCO2e per declared unit"
    )
    pct60_gwp: float = pyd.Field(
        description="60th percentile GWP for this statistics measured in kgCO2e per declared unit"
    )
    pct70_gwp: float = pyd.Field(
        description="70th percentile GWP for this statistics measured in kgCO2e per declared unit"
    )
    conservative_estimate: float = pyd.Field(
        description="Conservative estimate. 80th percentile of GWP per declared unit measured in kgCO2e",
        example=640.778,
    )
    pct90_gwp: float = pyd.Field(
        description="70th percentile GWP for this statistics measured in kgCO2e per declared unit"
    )

    # stats
    average: float = pyd.Field(description="Average GWP in kgCO2e per declared unit", example=554.2)
    min: float | None = pyd.Field(
        description="Min GWP of returned results measured in kgCO2e per declared unit", example=998.3
    )
    max: float | None = pyd.Field(
        description="Max GWP of returned results measured in kgCO2e per declared unit", example=120.0
    )

    # percentiles w/out burden of doubt
    pct20_gwp_no_bod: float | None = pyd.Field(
        description="20th percentile of GWP (kgCO2e per declared unit), no burden of doubt", example=120
    )
    pct40_gwp_no_bod: float | None = pyd.Field(
        description="40th percentile of GWP (kgCO2e per declared unit), no burden of doubt", example=120
    )
    pct60_gwp_no_bod: float | None = pyd.Field(
        description="60th percentile of GWP (kgCO2e per declared unit), no burden of doubt", example=120
    )
    pct80_gwp_no_bod: float | None = pyd.Field(
        description="80th percentile of GWP (kgCO2e per declared unit), no burden of doubt", example=120
    )
    average_gwp_no_bod: float | None = pyd.Field(description="Average GWP, no burden of doubt", example=120)

    # set parameters
    standard_deviation: float = pyd.Field(description="Standard deviation", example=87.62)
    epds_count: int = pyd.Field(description="Number of EPDs participated in statistics", example=55)
    industry_epds_count: int = pyd.Field(
        description="Number of Industry-wide EPDs participated in statistics", example=4
    )
    generic_estimates_count: int = pyd.Field(
        description="Number of Generic Estimates participated in statistics", example=0
    )

    declared_unit: Amount = pyd.Field(
        description="Declared unit for the statistics. "
        "Statistical values - percentiles, averages etc - are based on this unit of product"
    )


class EpdSearchMeta(MaterialFilterMetaMixin, PagingMetaMixin, WarningMetaMixin, BaseMeta):
    """Metadata for EPD Search endpoint."""

    pass


class EpdStatisticsMeta(MaterialFilterMetaMixin, WarningMetaMixin, BaseMeta):
    """Metadata for EPD Statistics endpoint."""

    pass


EpdSearchResponse: TypeAlias = OpenEpdApiResponse[list[Epd], EpdSearchMeta]
EpdStatisticsResponse: TypeAlias = OpenEpdApiResponse[StatisticsDto, EpdStatisticsMeta]



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import (
    AhuAirflowControl,
    AhuZoneControl,
    AirFiltersMediaType,
    AirFiltersMervRating,
    BoilerConfiguration,
    BoilerEquipmentFuelType,
    HeatPumpType,
    HvacDuctMaterial,
    HvacDuctShape,
    HvacDuctType,
    HvacHeatExchangersType,
    MechanicalInstallation,
    MechanicalRefrigerants,
)
from openepd.model.validation.quantity import AirflowStr, FlowRateStr, PowerStr, PressureMPaStr, VolumeStr


class HvacVrfControlV1(BaseOpenEpdHierarchicalSpec):
    """Controller for adjusting airflow across the VRF system."""

    _EXT_VERSION = "1.0"


class HvacVrfIndoorV1(BaseOpenEpdHierarchicalSpec):
    """Heating and cooling unit located on the inside of a building and supplies air to specific indoor zones."""

    _EXT_VERSION = "1.0"

    # Own fields:
    refrigerants: list[MechanicalRefrigerants] | None = pyd.Field(default=None, description="", example=["R11"])
    heating_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    cooling_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    airflow_rate: AirflowStr | None = pyd.Field(default=None, description="", example="1 m3 / s")
    air_volume: VolumeStr | None = pyd.Field(default=None, description="", example="1 m3")


class HvacVrfOutdoorV1(BaseOpenEpdHierarchicalSpec):
    """Heating and cooling unit that is on the outside of a building and distributes air to the indoor units."""

    _EXT_VERSION = "1.0"

    # Own fields:
    refrigerants: list[MechanicalRefrigerants] | None = pyd.Field(default=None, description="", example=["R11"])
    heating_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    cooling_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    airflow_rate: AirflowStr | None = pyd.Field(default=None, description="", example="1 m3 / s")
    air_volume: VolumeStr | None = pyd.Field(default=None, description="", example="1 m3")


class HvacAirDiffusersV1(BaseOpenEpdHierarchicalSpec):
    """Room-side terminals for air distribution. This is different from Terminal Heating & Cooling Units."""

    _EXT_VERSION = "1.0"


class HvacAirFiltersV1(BaseOpenEpdHierarchicalSpec):
    """Device for filtering particles of dust, soot, etc., from the air passing through it."""

    _EXT_VERSION = "1.0"

    # Own fields:
    merv_rating: AirFiltersMervRating | None = pyd.Field(default=None, description="", example="MERV 1")
    media_type: AirFiltersMediaType | None = pyd.Field(default=None, description="", example="Acrylic")


class HvacAHUsV1(BaseOpenEpdHierarchicalSpec):
    """
    Device which provides healthy, dust free air to buildings with a good energy efficiency.

    Usually a large metal box containing a blower, heating or cooling elements, filter racks or chambers, sound
    attenuators, and dampers.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    refrigerants: list[MechanicalRefrigerants] | None = pyd.Field(default=None, description="", example=["R11"])
    installation: MechanicalInstallation | None = pyd.Field(default=None, description="", example="Indoor")
    airflow_rate: AirflowStr | None = pyd.Field(default=None, description="", example="1 m3 / s")
    air_volume: VolumeStr | None = pyd.Field(default=None, description="", example="1 m3")
    cooling_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    heating_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    airflow_control: AhuAirflowControl | None = pyd.Field(default=None, description="", example="CAV")
    zone_control: AhuZoneControl | None = pyd.Field(default=None, description="", example="Single Zone")


class HvacBoilersV1(BaseOpenEpdHierarchicalSpec):
    """Closed vessel for heating fluid."""

    _EXT_VERSION = "1.0"

    # Own fields:
    flow_rate: FlowRateStr | None = pyd.Field(default=None, description="", example="1 l / min")
    heating_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    configuration: BoilerConfiguration | None = pyd.Field(default=None, description="", example="Hot water")
    fuel_type: BoilerEquipmentFuelType | None = pyd.Field(default=None, description="", example="Coal")


class HvacChillersV1(BaseOpenEpdHierarchicalSpec):
    """
    Machine that removes heat from a liquid coolant.

    Uses a vapor-compression, adsorption refrigeration, or absorption refrigeration cycles. Incldues centrifugal,
    water-cooled chillers, etc.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    refrigerants: list[MechanicalRefrigerants] | None = pyd.Field(default=None, description="", example=["R11"])
    installation: MechanicalInstallation | None = pyd.Field(default=None, description="", example="Indoor")
    heating_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    cooling_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    air_volume: VolumeStr | None = pyd.Field(default=None, description="", example="1 m3")
    airflow_rate: AirflowStr | None = pyd.Field(default=None, description="", example="1 m3 / s")


class HvacFansV1(BaseOpenEpdHierarchicalSpec):
    """Apparatus with rotating blades that creates a current of air for cooling or ventilation."""

    _EXT_VERSION = "1.0"


class HvacHeatPumpsV1(BaseOpenEpdHierarchicalSpec):
    """Device that transfers thermal energy between spaces, including ground and air source heat pumps."""

    _EXT_VERSION = "1.0"

    # Own fields:
    refrigerants: list[MechanicalRefrigerants] | None = pyd.Field(default=None, description="", example=["R11"])
    cooling_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    heating_capacity: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")
    air_volume: VolumeStr | None = pyd.Field(default=None, description="", example="1 m3")
    airflow_rate: AirflowStr | None = pyd.Field(default=None, description="", example="1 m3 / s")
    heat_pumps_type: HeatPumpType | None = pyd.Field(default=None, description="", example="Air-to-Water")


class HvacHeatExV1(BaseOpenEpdHierarchicalSpec):
    """
    HVAC heat exchangers.

    Systems that with heat exchange cell which recovers and retains the heat that would otherwise be lost from the
    extracted air.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    refrigerants: list[MechanicalRefrigerants] | None = pyd.Field(default=None, description="", example=["R11"])
    heat_exchangers_type: HvacHeatExchangersType | None = pyd.Field(
        default=None, description="", example="Shell and Tube"
    )


class HvacPumpsV1(BaseOpenEpdHierarchicalSpec):
    """
    Pumps.

    Mechanical device using suction or pressure to raise or move liquids, compress gases, or force air into
    inflatable objects such as tires.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    flow_rate: FlowRateStr | None = pyd.Field(default=None, description="", example="1 l / min")
    pump_discharge_pressure: PressureMPaStr | None = pyd.Field(default=None, description="", example="1 MPa")
    pump_horsepower: PowerStr | None = pyd.Field(default=None, description="", example="1000.0 W")


class HvacRTUsV1(BaseOpenEpdHierarchicalSpec):
    """An air handler designed for outdoor use, typically on roofs."""

    _EXT_VERSION = "1.0"


class HvacVrfSystemsV1(BaseOpenEpdHierarchicalSpec):
    """
    Variable refrigerant flow (VRF).

    Also known as variable refrigerant volume (VRV), is an HVAC technology that allows for varying degrees of cooling
    in more specific areas.
    """

    _EXT_VERSION = "1.0"

    # Nested specs:
    HvacVrfControl: HvacVrfControlV1 | None = None
    HvacVrfIndoor: HvacVrfIndoorV1 | None = None
    HvacVrfOutdoor: HvacVrfOutdoorV1 | None = None


class HvacDuctsV1(BaseOpenEpdHierarchicalSpec):
    """Ducts for HVAC systems performance specification."""

    _EXT_VERSION = "1.0"

    shape: HvacDuctShape | None = pyd.Field(default=None, description="Hvac duct shape", example="Rectangular")
    material: HvacDuctMaterial | None = pyd.Field(
        default=None, description="Hvac duct material", example="Galvanized Steel"
    )
    type: HvacDuctType | None = pyd.Field(default=None, description="Hvac duct type", example="Flexible")


class MechanicalV1(BaseOpenEpdHierarchicalSpec):
    """Mechanical performance specification."""

    _EXT_VERSION = "1.1"

    # Nested specs:
    HvacAirDiffusers: HvacAirDiffusersV1 | None = None
    HvacAirFilters: HvacAirFiltersV1 | None = None
    HvacAHUs: HvacAHUsV1 | None = None
    HvacBoilers: HvacBoilersV1 | None = None
    HvacChillers: HvacChillersV1 | None = None
    HvacFans: HvacFansV1 | None = None
    HvacHeatPumps: HvacHeatPumpsV1 | None = None
    HvacHeatEx: HvacHeatExV1 | None = None
    HvacPumps: HvacPumpsV1 | None = None
    HvacRTUs: HvacRTUsV1 | None = None
    HvacVrfSystems: HvacVrfSystemsV1 | None = None
    HvacDucts: HvacDuctsV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "AuxiliariesRangeV1",
    "CleaningProductsRangeV1",
    "ClothingRangeV1",
    "FoodBeverageRangeV1",
    "TransportationInfrastructureRangeV1",
    "UnsupportedRangeV1",
    "CopperRangeV1",
    "EarthworkRangeV1",
    "ExteriorSunControlDevicesRangeV1",
    "GypsumFinishingCompoundsRangeV1",
    "ProfilesRangeV1",
    "UnknownRangeV1",
    "ZincRangeV1",
    "OtherMaterialsRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec


class AuxiliariesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Auxiliaries performance specification.

    Range version.
    """

    _EXT_VERSION = "1.0"


class CleaningProductsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Cleaning and disinfecting solutions.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ClothingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Clothing.

    Range version.
    """

    _EXT_VERSION = "1.0"


class FoodBeverageRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Food Beverage and Tobacco Products.

    Range version.
    """

    _EXT_VERSION = "1.0"


class TransportationInfrastructureRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    A broad category for unclassified materials focused on transportation infrastructure.

    Range version.
    """

    _EXT_VERSION = "1.0"

    Auxiliaries: AuxiliariesRangeV1 | None = None


class UnsupportedRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    A generic category for EPDs/Materials and categories that are explicitly not yet supported by EC3.

    Assume that any data in this subcategory is unreliable.

    Range version.
    """

    _EXT_VERSION = "1.0"

    CleaningProducts: CleaningProductsRangeV1 | None = None
    Clothing: ClothingRangeV1 | None = None
    FoodBeverage: FoodBeverageRangeV1 | None = None


class CopperRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Products made of copper.

    Range version.
    """

    _EXT_VERSION = "1.0"


class EarthworkRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Earthwork, including excavation, shoring, piles, etc.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ExteriorSunControlDevicesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Sun control devices help to manage solar heat gain by redirecting sunlight.

    Range version.
    """

    _EXT_VERSION = "1.0"


class GypsumFinishingCompoundsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Plasters and the like for finishing Gypsum Sheet and Board.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ProfilesRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Metal or polymer profiles used for producing or installing of Windows, Doors, Frames and Cladding.

    Range version.
    """

    _EXT_VERSION = "1.0"


class UnknownRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Materials with unknown category. Assume that any data in this subcategory is unreliable.

    Range version.
    """

    _EXT_VERSION = "1.0"


class ZincRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Products made of zinc.

    Range version.
    """

    _EXT_VERSION = "1.0"


class OtherMaterialsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Broad category of materials not yet classified.

    Range version.
    """

    _EXT_VERSION = "1.0"

    TransportationInfrastructure: TransportationInfrastructureRangeV1 | None = None
    Unsupported: UnsupportedRangeV1 | None = None
    Copper: CopperRangeV1 | None = None
    Earthwork: EarthworkRangeV1 | None = None
    ExteriorSunControlDevices: ExteriorSunControlDevicesRangeV1 | None = None
    GypsumFinishingCompounds: GypsumFinishingCompoundsRangeV1 | None = None
    Profiles: ProfilesRangeV1 | None = None
    Unknown: UnknownRangeV1 | None = None
    Zinc: ZincRangeV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Any

from pydantic import utils as pydantic_utils

from openepd.compat.pydantic import pyd


def patch_pydantic_metaclass():
    """
    Patch pydantic's ModelMetaclass to restore class attribute lookup for fields.

    In pydantic, while the model fields are defined in the model as class-level attributes, in runtime they disappear
    due to ModelMetaclass logic. ModelMetaclass takes the defined attributes, removes them from class dict and puts
    into a special __fields__ attribute to avoid naming conflict.

    We would like to be able to access the attributes via dot notation in the runtimes, since it makes refactoring
    easier.

    This class exposes the original fields when accessed via class name. For example, one can call `Pcr.name` and get
    `ModelField`, in addition to calling `pcr.__fields__` on an instance.
    """

    def model_metaclass__getattr__(cls, name: str) -> Any:
        if name in cls.__fields__:
            return cls.__fields__[name]
        return getattr(super, name)

    pyd.main.ModelMetaclass.__getattr__ = model_metaclass__getattr__


def patch_pydantic_metaclass_validator():
    """
    Patch the internal validator function used during model construction to support modified metaclass.

    Pydantic has a special guard which stops execution if a model defines field which shadows a basemodel interface.
    For example, if someone would define a field named `__fields__` this would break code.

    With the modified metaclass functionality, we are exposing the original fields as class attributes, and this
    breaks this check.

    This patcher method modifies the pydantic internals so that the check is retained, but it is not causing exception
    when doing the normal field inheritance.
    """
    model_field_classes = list()

    try:
        from pydantic.v1.fields import ModelField as ModelFieldV1

        model_field_classes.append(ModelFieldV1)
    except ImportError:
        pass

    try:
        from pydantic.fields import ModelField as ModelFieldPydanticV1

        model_field_classes.append(ModelFieldPydanticV1)
    except ImportError:
        pass

    try:
        from pydantic_core.core_schema import ModelField as ModelFieldV2

        model_field_classes.append(ModelFieldV2)
    except ImportError:
        pass
    model_field_classes_tuple = tuple(model_field_classes)

    def pydantic_utils__validate_field_name(bases: list[type[pyd.BaseModel]], field_name: str) -> None:
        for base in bases:
            if attr := getattr(base, field_name, None):
                if isinstance(attr, model_field_classes_tuple):
                    continue

                raise NameError(
                    f'Field name "{field_name}" shadows a BaseModel attribute; '
                    f"use a different field name with \"alias='{field_name}'\"."
                )

    pyd.main.validate_field_name = pydantic_utils__validate_field_name
    pydantic_utils.validate_field_name = pydantic_utils__validate_field_name


def patch_pydantic():
    """
    Modify Pydantic to support field attribute access via class.

    Example: Field the_field in the model TheModel(BaseModel). Before this patch, one can do
     `TheModel().__fields__["the_field"]' to get field descriptor. After the fix, one can do TheModel.the_field.

     To disable, set env variable `OPENEPD_DISABLE_PYDANTIC_PATCH` to "1", "yes" or "true".
    """
    patch_pydantic_metaclass()
    patch_pydantic_metaclass_validator()
